from deltalake import Field, Schema
from deltalake.schema import ArrayType
from json import loads
from polars.datatypes import (
  Boolean,
  DataType as plDataType,
  Date,
  Datetime,
  Float32,
  Float64,
  Int32,
  Int64,
  List,
  String
)
from typing import Optional, Union

from polta.enums import TableQuality
from polta.exceptions import DataTypeNotRecognized
from polta.schemas.table import (
  raw_metadata,
  conformed_metadata,
  canonical_metadata
)


class Maps:
  """Contains various mapper fields and methods for Polta operations"""
  DELTALAKE_TO_POLARS_FIELD: dict[str, plDataType] = {
    'boolean': Boolean,
    'date': Date,
    'double': Float64,
    'float': Float32,
    'integer': Int32,
    'long': Int64,
    'string': String,
    'timestamp': Datetime(time_zone='UTC')
  }
  POLARS_TO_DELTALAKE_FIELD: dict[plDataType, str] = {
    Boolean: 'boolean',
    Date: 'date',
    Float64: 'double',
    Float32: 'float',
    Int32: 'integer',
    Int64: 'long',
    String: 'string',
    Datetime(time_zone='UTC'): 'timestamp'
  }
  QUALITY_TO_METADATA_COLUMNS: dict[str, list[Field]] = {
    TableQuality.RAW.value: raw_metadata.fields,
    TableQuality.CONFORMED.value: conformed_metadata.fields,
    TableQuality.CANONICAL.value: canonical_metadata.fields
  }
  QUALITY_TO_FAILURE_COLUMN: dict[str, str] = {
    TableQuality.RAW.value: '_raw_id',
    TableQuality.CONFORMED.value: '_conformed_id',
    TableQuality.CANONICAL.value: '_canonicalized_id'
  }

  @staticmethod
  def deltalake_field_to_polars_field(delta_field: Union[dict, str]) -> plDataType:
    """Maps an individual Delta field to a Polars data type
    
    Args:
      delta_field (str): the field name (e.g., 'boolean', 'date')
    
    Returns:
      dt (DataType): the resulting Polars data type
    """
    wrap_in_list: bool = False
    if isinstance(delta_field, dict) and delta_field.get('type') == 'array':
      delta_field: str = delta_field.get('elementType')
      wrap_in_list: bool = True
    if not isinstance(delta_field, str):
      raise TypeError('Error: delta_field must be of type <str> or <dict>')
    dt: Union[plDataType, str] = Maps.DELTALAKE_TO_POLARS_FIELD.get(delta_field, '')
    if isinstance(dt, str):
      raise DataTypeNotRecognized(dt)
    return List(dt) if wrap_in_list else dt

  @staticmethod
  def deltalake_schema_to_polars_schema(schema: Schema) -> dict[str, plDataType]:
    """Converts the existing Delta Lake schema and returns it as a dict compatible to Polars DataFrames

    Args:
      schema (Schema): the original deltalake schema
    
    Returns:
      schema.polars (dict[str, DataType]): the schema as a dict, compatible with Polars DataFrames
    """
    polars_schema: dict[str, plDataType] = {}

    for field in loads(schema.to_json())['fields']:
      polars_schema[field['name']] = Maps.deltalake_field_to_polars_field(field['type'])
      if field['type'] == 'timestamp':
        polars_schema[field['name']]
      
    return polars_schema  

  @staticmethod
  def polars_field_to_deltalake_field(column: str, data_type: plDataType) -> Field:
    """Converts a polars field to a deltalake field
    
    Args:
      column (str): the name of the column
      data_type (DataType): polars DataType
    
    Returns:
      field (Field): the resulting deltalake field
    """
    try:
      if isinstance(data_type, List):
        dt: Optional[str] = Maps.POLARS_TO_DELTALAKE_FIELD[data_type.inner]
        return Field(column, ArrayType(dt))
      else:
        dt: Optional[str] = Maps.POLARS_TO_DELTALAKE_FIELD[data_type]
        return Field(column, dt)
    except KeyError:
      raise DataTypeNotRecognized(data_type)
  
  @staticmethod
  def polars_schema_to_deltalake_schema(schema: dict[str, plDataType]) -> Schema:
    """Converts a polars schema to a deltalake schema
    
    Args:
      schema (dict[str, DataType]): the original polars schema
    
    Returns:
      schema (Schema): the resulting deltalake schema
    """
    fields: list[plDataType] = []
    for column, data_type in schema.items():
      fields.append(Maps.polars_field_to_deltalake_field(column, data_type))
    return Schema(fields)

  @staticmethod
  def quality_to_failure_column(quality: TableQuality) -> str:
    """Converts a table quality to the failure column
    
    Args:
      quality (TableQuality): the quality of the table
    
    Returns:
      failure_column (str): the name of the failure column for that quality
    """
    return Maps.QUALITY_TO_FAILURE_COLUMN[quality.value]

  @staticmethod
  def quality_to_metadata_columns(quality: TableQuality) -> list[Field]:
    """Converts a table quality to metadata columns for that quality
  
    Args:
      quality (TableQuality): the quality of the table
    
    Returns:
      metadata_columns(list[Field]): the metadata columns for that quality
    """
    return Maps.QUALITY_TO_METADATA_COLUMNS[quality.value]



import polars as pl

from dataclasses import dataclass, field
from datetime import datetime, UTC
from deltalake import DeltaTable, Schema, TableFeatures
from os import getcwd, listdir, makedirs, path
from polars import DataFrame
from typing import Any

from polta.enums import TableQuality
from polta.exceptions import DomainDoesNotExist
from polta.schemas.system import file_history, pipe_history


@dataclass
class Metastore:
  """Dataclass for managing Polta metastores
  
  Optional Args:
    main_path (str): the directory of the metastore (default CWD + 'metastore')

  Initialized Fields:
    tables_directory (str): the path to the tables
    volumes_directory (str): the path to the volumes
  """
  main_path: str = field(default_factory=lambda: path.join(getcwd(), 'metastore'))

  tables_directory: str = field(init=False)
  volumes_directory: str = field(init=False)
  file_history_path: str = field(init=False)
  pipe_history_path: str = field(init=False)

  def __post_init__(self) -> None:
    self.tables_directory: str = path.join(self.main_path, 'tables')
    self.volumes_directory: str = path.join(self.main_path, 'volumes')
    self.sys_directory: str = path.join(self.volumes_directory, 'system')
    self.file_history_path: str = path.join(self.sys_directory, 'file_history')
    self.pipe_history_path: str = path.join(self.sys_directory, 'pipe_history')
    self.initialize_if_not_exists()

  def initialize_if_not_exists(self) -> None:
    """Initializes the metastore if it does not exist"""
    # Initialize the directories
    makedirs(self.main_path, exist_ok=True)
    makedirs(self.tables_directory, exist_ok=True)
    makedirs(self.volumes_directory, exist_ok=True)

    # Initialize the system tables
    self.create_table_if_not_exists(
      table_path=self.file_history_path,
      schema=file_history,
      partition_by=['table_id']
    )
    self.create_table_if_not_exists(
      table_path=self.pipe_history_path,
      schema=pipe_history,
      partition_by=['pipe_id']
    )

  @staticmethod
  def create_table_if_not_exists(table_path: str, schema: Schema,
                                 partition_by: list[str] = []) -> None:
    """Creates a Delta Table if it does not exist

    Args:
        table_path (str): the path of the Delta Table
        schema (Schema): the table schema, in case the Delta Table needs to be created
        partition_by (list[str]): if applicable, the list of keys by which to partition
    """
    if not isinstance(table_path, str):
      raise TypeError('Error: table_path must be of type <str>')
    if not isinstance(schema, Schema):
      raise TypeError('Error: schema must be of type <Schema>')
    if not isinstance(partition_by, list):
      raise TypeError('Error: partition_by must be of type <list>')
    if not all(isinstance(k, str) for k in partition_by):
      raise TypeError('Error: all values in partition_by must be of type <str>')

    # If it exists already, return
    if DeltaTable.is_deltatable(table_path):
      return

    dt: DeltaTable = DeltaTable.create(
      table_uri=table_path,
      schema=schema,
      mode='ignore',
      partition_by=partition_by or None
    )
    dt.alter.add_feature(
      feature=TableFeatures.TimestampWithoutTimezone,
      allow_protocol_versions_increase=True
    )

  def list_domains(self) -> list[str]:
    """Retrieves the directories and names of available domains
    
    Returns:
      domains (list[str]): the available qualities
    """
    return listdir(self.tables_directory)

  def list_qualities(self, domain: str) -> list[TableQuality]:
    """Retrieves the available table qualities for a domain
    
    Args:
      domain (str): the domain to check
    
    Returns:
      qualities (list[TableQuality]): the available qualities for that domain
    """
    qualities_path: str = path.join(self.tables_directory, domain)
    if not path.exists(qualities_path):
      raise DomainDoesNotExist(domain)
    return [TableQuality(q) for q in listdir(qualities_path)]

  def domain_exists(self, domain: str) -> bool:
    """Indicates whether the domain exists
    
    Args:
      domain (str): the domain to check
    
    Returns:
      domain_exists (bool): indicates whether the domain exists
    """
    return path.exists(path.join(self.tables_directory, domain))

  def quality_exists(self, domain: str, quality: TableQuality) -> bool:
    """Indicates whether the quality exists under a given domain
    
    Args:
      domain (str): the domain containing the quality to check
      quality (TableQuality): the quality to check
    
    Returns:
      quality_exists (bool): indicates whether the quality exists
    """
    return path.exists(path.join(self.tables_directory, domain, quality.value))

  def get_file_history(self, table_id: str) -> DataFrame:
    """Retrieves a file_history DataFrame for a table by id

    Args:
      table_id (str): the unique ID of the table
    
    Returns:
      df (DataFrame): the resulting file history DataFrame
    """
    return (pl
      .read_delta(self.file_history_path)
      .filter(pl.col('table_id').eq(table_id))
      .select('_file_path', '_file_mod_ts')
      .unique()
      .group_by('_file_path')
      .agg(pl.col('_file_mod_ts').max())
    )

  def write_file_history(self, table_id: str, df: DataFrame) -> None:
    """Writes the file history into the system table
    
    Args:
      table_id (str): the unique ID of the table
      df (DataFrame): the DataFrame containing history information
    """
    df: DataFrame = (df
      .select(
        pl.lit(table_id).alias('table_id'),
        pl.col('_file_path'),
        pl.col('_file_mod_ts'),
        pl.col('_ingested_ts')
      )
    )
    df.write_delta(
      target=self.file_history_path,
      mode='append'
    )

  def clear_file_history(self, table_id: str) -> None:
    """Removes the file history of a table, typically after truncation
    
    Args:
      table_id (str): the unique ID of the table
    """
    DeltaTable(self.file_history_path).delete(f'table_id = \'{table_id}\'')

  def write_pipe_history(self, pipe_id: str, execution_start_ts: datetime, strict: bool,
                         succeeded: bool, in_memory: bool, passed_count: int,
                         failed_count: int, quarantined_count: int) -> None:
    """Writes a record to the pipe_history system table

    Args:
      pipe_id (str): the unique ID of the pipe
      execution_start_ts (datetime): when the pipe began
      strict (bool): indicates whether the pipe ran in strict mode
      succeeded (bool): indicates whether the pipe succeeded
      in_memory (bool): indicates whether the pipe ran in-memory
      passed_count (int): the number of records that passed
      failed_count (int): the number of records that failed
      quarantined_count (int): the number of records that got quarantined
    """
    now: datetime = datetime.now(UTC)
    record: dict[str, Any] = {
      'pipe_id': pipe_id,
      'execution_start_ts': execution_start_ts,
      'execution_end_ts': now,
      'execution_duration': (now - execution_start_ts).total_seconds(),
      'strict': strict,
      'succeeded': succeeded,
      'in_memory': in_memory,
      'total_count': passed_count + failed_count + quarantined_count,
      'passed_count': passed_count,
      'failed_count': failed_count,
      'quarantined_count': quarantined_count
    }
    DataFrame([record]).write_delta(
      target=self.pipe_history_path,
      mode='append'
    )

  def get_pipe_history(self, pipe_id: str = '') -> DataFrame:
    """Retrieves the pipe_history system table
    
    Args:
      pipe_id (str): if applicable, the unique ID of the pipe
    
    Returns:
      df (DataFrame): the pipe_history DataFrame
    """
    df: DataFrame = pl.read_delta(self.pipe_history_path)
    if pipe_id:
      df: DataFrame = df.filter(pl.col('pipe_id').eq(pipe_id))
    return df



import polars as pl

from dataclasses import dataclass, field
from datetime import datetime
from deltalake import DeltaTable, Schema, TableFeatures
from os import makedirs, path
from pathlib import Path
from polars import DataFrame
from shutil import rmtree
from typing import Optional

from polta.test import Test
from polta.enums import CheckAction, TableQuality
from polta.exceptions import PoltaDataFormatNotRecognized
from polta.metastore import Metastore
from polta.table_schema import TableSchema
from polta.types import RawPoltaData


@dataclass
class Table:
  """Stores all applicable information for a Polars + Delta Table dataset
  
  Positional Args:
    domain (str): the kind of data this table contains
    quality (TableQuality): the quality of the data
    name (str): the name of the table
    
  Optional Args:
    raw_schema (Optional[Schema]): the raw table schema (default None)
    metastore (Metastore): The metastore (default Metastore())
    primary_keys (list[str]): for upserts, the primary keys of the table (default [])
    tests (list[Test]): test checks before loading any data
  
  Initialized fields:
    id (str): the unique identifier for the table
    schema (TableSchema): the table schema object
    table_path (str): the absolute path to the Table in the metastore
    ingestion_zone_path (str): the path to the ingestion zone
    state_file_directory (str): the absolute path to the state files directory
    state_file_path (str): the absolute path to the Table state file
    schema.polars (dict[str, DataType]): the table schema as a Polars object
    schema.deltalake (Schema): the table schema as a deltalake object
    columns (list[str]): the table columns
    merge_predicate (str): the SQL merge predicate for upserts
    quarantine_path (str): the path to the corresponding quarantine table
    schema.quarantine (Schema): the schema of the corresponding quarantine table
    failure_column (str): the column name for identifying failure records
  """
  domain: str
  quality: TableQuality
  name: str
  raw_schema: Optional[Schema] = field(default_factory=lambda: None)
  metastore: Metastore = field(default_factory=lambda: Metastore())
  primary_keys: list[str] = field(default_factory=lambda: [])
  tests: list[Test] = field(default_factory=lambda: [])

  id: str = field(init=False)
  schema: TableSchema = field(init=False)
  table_path: str = field(init=False)
  ingestion_zone_path: str = field(init=False)
  state_file_directory: str = field(init=False)
  state_file_path: str = field(init=False)
  quarantine_path: str = field(init=False)
  merge_predicate: Optional[str] = field(init=False)

  def __post_init__(self) -> None:
    self.id: str = '.'.join([
      self.domain,
      self.quality.value,
      self.name
    ])
    self.schema: TableSchema = TableSchema(self.raw_schema, self.quality)
    self.table_path: str = path.join(
      self.metastore.tables_directory,
      self.domain,
      self.quality.value,
      self.name
    )
    self.ingestion_zone_path: str = path.join(
      self.metastore.volumes_directory,
      'ingestion',
      self.domain,
      self.name
    )
    self.state_file_directory: str = path.join(
      self.metastore.volumes_directory,
      'state',
      self.domain,
      self.quality.value,
      self.name            
    )
    self.state_file_path: str = path.join(
      self.state_file_directory,
      '.STATE'
    )
    self.quarantine_path: str = path.join(
      self.metastore.volumes_directory,
      'quarantine',
      self.domain,
      self.quality.value,
      self.name
    )

    if self.primary_keys:
      self.merge_predicate: Optional[str] = Table.build_merge_predicate(self.primary_keys)
    else:
      self.merge_predicate: Optional[str] = None
    if self.quality.value == TableQuality.RAW.value:
      self._build_ingestion_zone_if_not_exists()

  @staticmethod
  def create_if_not_exists(table_path: str, schema: Schema) -> None:
    """Creates a Delta Table if it does not exist

    Args:
        table_path (str): the path of the Delta Table
        schema (Schema): the table schema, in case the Delta Table needs to be created
    """
    if not isinstance(table_path, str):
      raise TypeError('Error: table_path must be of type <str>')
    if not isinstance(schema, Schema):
      raise TypeError('Error: schema must be of type <Schema>')

    # If it exists already, return
    if DeltaTable.is_deltatable(table_path):
      return

    dt: DeltaTable = DeltaTable.create(table_path, schema, mode='ignore')
    dt.alter.add_feature(
      feature=TableFeatures.TimestampWithoutTimezone,
      allow_protocol_versions_increase=True
    )   

  @staticmethod
  def build_merge_predicate(primary_keys: list[str]) -> str:
    """Constructs a merge predicate based on the source/target aliases and primary keys

    Args:
      primary_keys (list[str]): the primary keys for the merge

    Returns:
      merge_predicate (str): the merge predicate as a conjunction of SQL conditions matching on primary keys
    """
    return ' AND '.join([f's.{k} = t.{k}' for k in primary_keys])

  def enforce_dataframe(self, data: RawPoltaData) -> DataFrame:
    """Takes either a DataFrame or record(s) and returns the DataFrame representation
    
    Args:
      data (RawPoltaData): the data to enforce
    
    Returns:
      df (DataFrame): the DataFrame representation
    """
    if isinstance(data, dict):
      return DataFrame([data], self.schema.polars)
    elif isinstance(data, list) and all(isinstance(r, dict) for r in data):
      return DataFrame(data, self.schema.polars)
    elif isinstance(data, DataFrame):
      return data
    else:
      raise PoltaDataFormatNotRecognized(type(data))

  def apply_tests(self, df: DataFrame) -> tuple[DataFrame, DataFrame, DataFrame]:
    """Applies each test case to the data
    
    Args:
      df (DataFrame): the DataFrame to test
        
    Returns:
      passed, failed, quarantined (DataFrame): the resulting DataFrames
    """
    # Build three empty DataFrames to hold results
    passed: DataFrame = DataFrame([], self.schema.polars)
    failed: DataFrame = DataFrame([], self.schema.quarantine)
    quarantined: DataFrame = DataFrame([], self.schema.quarantine)

    # Skip all tests they do not exist
    if not self.tests:
      return df, failed, quarantined

    # Iterate the available tests
    for test in self.tests:
      # Execute the test against the original DataFrame or the remaining records
      res_df: DataFrame = test.run(df if passed.is_empty() else passed)

      # Store the results either in a passed DataFrame or failed DataFrame
      passed: DataFrame = res_df.filter(pl.col(test.result_column) == 1).drop(test.result_column)
      failed_test: DataFrame = (res_df
        .filter(pl.col(test.result_column) == 0)
        .drop(test.result_column)
        .with_columns(pl.lit(test.check.name).alias('failed_test'))
      )

      # Depending on the action, place the failed tests in the appropriate DataFrame
      if test.check_action.value == CheckAction.FAIL.value:
        failed: DataFrame = pl.concat([failed, failed_test])
      elif test.check_action.value == CheckAction.QUARANTINE.value:
        quarantined: DataFrame = pl.concat([quarantined, failed_test])

    # Return the resulting DataFrames
    return passed, failed, quarantined

  def truncate(self) -> None:
    """Truncates the table"""
    self.overwrite(DataFrame([], self.schema.polars))
    self.metastore.clear_file_history(self.id)

  def drop(self) -> None:
    """Drops the table"""
    if path.exists(self.table_path) and DeltaTable.is_deltatable(self.table_path):
      rmtree(self.table_path)

  def clear_quarantine(self) -> None:
    """Clears the quarantine table if it exists"""
    if path.exists(self.quarantine_path) and \
     DeltaTable.is_deltatable(self.quarantine_path):
      df: DataFrame = DataFrame([], self.schema.quarantine)
      df.write_delta(self.quarantine_path, mode='overwrite')

  def get_as_delta_table(self) -> DeltaTable:
    """Retrieves the DeltaTable object for the Table
    
    Returns:
      delta_table (DeltaTable): the resulting Delta Table
    """
    self.create_if_not_exists(self.table_path, self.schema.deltalake)
    return DeltaTable(self.table_path)

  def get(self, filter_conditions: dict = {}, partition_by: list[str] = [], order_by: list[str] = [],
          order_by_descending: bool = True, select: list[str] = [], sort_by: list[str] = [], limit: int = 0,
          unique: bool = False) -> DataFrame:
    """Retrieves a record, or records, by a specific condition, expecting only one record to return
      
    Args:
      filter_conditions (optional) (dict): if applicable, the filter conditions (e.g., {file_path: 'path.json'})
      partition_by (optional) (list[str]): if applicable, the keys by which to partition during deduplication
      order_by (optional) (list[str]): if applicable, the columns by which to order during deduplication
      order_by_descending (optional) (bool): if applicable, whether to ORDER BY DESC
      select (optional) (list[str]): if applicable, the columns to return after retrieving the DataFrame
      sort_by (optional) (list[str]): if applicable, the columns by which to sort the output
      limit (optional) (int): if applicable, a limit to the number of rows to return
      unique (optional) (bool): if applicable, remove any duplicate records

    Returns:
      df (DataFrame): the resulting DataFrame
    """
    if not isinstance(filter_conditions, dict):
      raise TypeError('Error: filter_conditions must be of type <dict>')
    if not isinstance(partition_by, list):
      raise TypeError('Error: partition_by must be of type <list>')
    if not isinstance(order_by, list):
      raise TypeError('Error: order_by must be of type <list>')
    if not isinstance(order_by_descending, bool):
      raise TypeError('Error: order_by_descending must be of type <bool>')
    if not isinstance(select, list):
      raise TypeError('Error: select must be of type <list>')
    if not isinstance(sort_by, list):
      raise TypeError('Error: sort_by must be of type <list>')
    if not isinstance(limit, int):
      raise TypeError('Error: limit must be of type <int>')
    if not isinstance(unique, bool):
      raise TypeError('Error: unique must be of type <bool>')
    if not all(isinstance(c, str) for c in partition_by):
      raise TypeError('Error: all values in partition_by must be of type <str>')
    if not all(isinstance(c, str) for c in order_by):
      raise TypeError('Error: all values in order_by must be of type <str>')
    if not all(isinstance(c, str) for c in select):
      raise TypeError('Error: all values in select must be of type <str>')
    if not all(isinstance(c, str) for c in sort_by):
      raise TypeError('Error: all values in sort_by must be of type <str>')

    # Create the Delta Table if it does not exist
    self.create_if_not_exists(self.table_path, self.schema.deltalake)

    # Retrieve Delta Table as a Polars DataFrame
    df: DataFrame = pl.read_delta(self.table_path)

    # Apply the filter condition if applicable
    if filter_conditions:
      df: DataFrame = df.filter(**filter_conditions)

    # Filter columns if applicable
    if select:
      df: DataFrame = df.select(select)

    # Apply a simple deduplication if applicable        
    if partition_by and order_by:
      df: DataFrame = df \
        .sort(order_by, descending=order_by_descending) \
        .unique(subset=partition_by, keep='first')
    
    # Apply a limit if applicable
    if limit:
      df: DataFrame = df.limit(limit)
    
    # Remove duplicate records if applicable
    if unique:
      df: DataFrame = df.unique()
      
    # Sort the results if applicable
    if sort_by:
      df: DataFrame = df.sort(sort_by)

    return df
  
  def upsert(self, data: RawPoltaData) -> None:
    """Upserts a DataFrame into the Delta Table

    self.primary_keys must be specified for this method to run
    
    Args:
      data (RawPoltaData): the data to upsert
    """
    if not self.primary_keys:
      raise ValueError('Error: Delta Table does not have primary keys')

    # Ensure table exists first
    self.create_if_not_exists(self.table_path, self.schema.deltalake)

    # Ensure DataFrame type
    df: DataFrame = self.enforce_dataframe(data)

    # Merge the DataFrame into its respective Delta Table
    # This merge logic is a simple upsert based on the table's primary keys
    (df
      .write_delta(
        target=self.table_path,
        mode='merge',
        delta_merge_options={
          'predicate': self.merge_predicate,
          'source_alias': 's',
          'target_alias': 't',
        }
      )
      .when_matched_update_all()
      .when_not_matched_insert_all()
      .execute()
    )
    self.touch_state_file()

  def overwrite(self, data: RawPoltaData) -> None:
    """Overwrites the Delta Table with the inputted DataFrame
    
    Args:
      data (RawPoltaData): the data with which to overwrite
    """
    # Ensure table exists first
    self.create_if_not_exists(self.table_path, self.schema.deltalake)

    # Ensure DataFrame type
    df: DataFrame = self.enforce_dataframe(data)

    df.write_delta(
      target=self.table_path,
      mode='overwrite'
    )
    self.touch_state_file()

  def append(self, data: RawPoltaData) -> None:
    """Appends a DataFrame to the Delta Table

    Args:
      data (RawPoltaData): the data with which to append
    """
    # Ensure table exists first
    self.create_if_not_exists(self.table_path, self.schema.deltalake)

    # Ensure DataFrame type
    df: DataFrame = self.enforce_dataframe(data)

    df.write_delta(
      target=self.table_path,
      mode='append'
    )
    self.touch_state_file()

  def touch_state_file(self) -> None:
    """Touches the state file to update the last modified datetime"""
    try:
      Path(self.state_file_path).touch()
    except FileNotFoundError:
      makedirs(self.state_file_directory, exist_ok=True)        
      Path(self.state_file_path).touch()

  def get_last_modified_datetime(self) -> datetime:
    """Retrieves the last modified datetime of the table
    
    Returns:
      last_modified_datetime (datetime): the last modified datetime of the table
    """
    modified_time: float = path.getmtime(self.state_file_path)
    return datetime.fromtimestamp(modified_time)

  def get_last_modified_record(self) -> dict[str, any]:
    """Builds a dict record for last modified information

    Returns:
      last_modified_record (dict[str, any]): the metadata of the table's last modified timestamp
    """
    return {
      'domain': self.domain,
      'quality': self.quality.value,
      'table': self.name,
      'path': self.table_path,
      'last_modified_datetime': self.get_last_modified_datetime()
    }

  def _build_ingestion_zone_if_not_exists(self) -> None:
    """Builds an empty directory for ingesting files"""
    if not path.exists(self.ingestion_zone_path):
      makedirs(self.ingestion_zone_path, exist_ok=True)
      print(f'Ingestion zone created: {self.ingestion_zone_path}')



from dataclasses import dataclass, field
from inspect import signature


@dataclass
class Check:
  """Contains check metadata and a function to generate results
  
  Args:
    name (str): the name of the check
    description (str): the description of the check
    function (callable): the polars transformation code to run the check
  
  Initialized Fields:
    simple_function (bool): indicates whether the arguments are simple (i.e., df, column)
  """
  name: str
  description: str
  function: callable

  simple_function: bool = field(init=False)

  def __post_init__(self) -> None:
    self.simple_function: bool = len(signature(self.function).parameters) == 2

  def build_result_column(self, column: str) -> str:
    """Builds the result column name based on an input column
    
    Args:
      column (str): the column that was checked
    
    Returns:
      result_column (str): the result column name
    """
    return f'__{column}__{self.name}__'



import polars as pl

from dataclasses import dataclass, field
from datetime import datetime, UTC
from deltalake import DeltaTable
from polars import DataFrame
from typing import Optional, Union
from uuid import uuid4

from polta.enums import WriteLogic, TableQuality
from polta.exceptions import (
  EmptyPipe,
  TableQualityNotRecognized,
  WriteLogicNotRecognized
)
from polta.exporter import Exporter
from polta.ingester import Ingester
from polta.table import Table
from polta.transformer import Transformer


@dataclass
class Pipe:
  """Changes and moves data in the metastore
  
  Positional Args:
    logic (Union[Ingester, Exporter, Transformer]): the pipe logic to handle data
  
  Initialized fields:
    id (str): the unique ID of the pipe for the pipeline
    table (Table): the destination Table
    write_logic (Optional[WriteLogic]): how the data should be placed in target table
  """
  logic: Union[Exporter, Ingester, Transformer]

  id: str = field(init=False)
  table: Table = field(init=False)
  write_logic: Optional[WriteLogic] = field(init=False)

  def __post_init__(self) -> None:
    self.id: str = '.'.join([
      'pp',
      self.logic.pipe_type.value,
      self.logic.table.id
    ])
    self.table: Table = self.logic.table
    self.write_logic = self.logic.write_logic

  def execute(self, dfs: dict[str, DataFrame] = {}, in_memory: bool = False,
              strict: bool = False) -> tuple[DataFrame, DataFrame, DataFrame]:
    """Executes the pipe

    Args:
      dfs (dict[str, DataFrame]): if applicable, source DataFrames (default {})
      in_memory (bool): indicates whether to run without saving (default False)
      strict (bool): indicates whether to fail on empty result (default False)

    Returns:
      passed, failed, quarantined (tuple[DataFrame, DataFrame, DataFrame]): the resulting DataFrames
    """
    print(f'Executing pipe {self.id}')
    
    # Record when the execution began
    execution_start: datetime = datetime.now(UTC)

    # Load in any extra data before transformation
    dfs.update(self.logic.get_dfs())

    # For in-memory exports, just carry over the table data
    # Otherwise, run the transformation/pre-load steps
    if isinstance(self.logic, Exporter) and in_memory:
      df: DataFrame = dfs[self.table.id]
    else:
      df: DataFrame = self.logic.transform(dfs)
      df: DataFrame = self.add_metadata_columns(df)
      df: DataFrame = self.conform_schema(df)

    # Run any tests and return the three data results
    passed, failed, quarantined = self.table.apply_tests(df)

    # Handle any quarantined records
    if not quarantined.is_empty():
      self.quarantine(quarantined)

    # If strict mode is enabled and dataset is empty, raise EmptyPipe
    succeeded: bool = (not strict) or (not passed.is_empty())

    # For standard runs and non-exports, save the passed data
    if isinstance(self.logic, (Ingester, Transformer)) and not in_memory:
      self.save(passed)
    
    # For exports, export the data
    if isinstance(self.logic, Exporter):
      self.logic.export(passed)

    # Print results
    print(f'  - Records passed: {passed.shape[0]}')
    print(f'  - Records failed: {failed.shape[0]}')
    print(f'  - Records quarantined: {quarantined.shape[0]}')

    # Log the pipe execution in the system table
    self.table.metastore.write_pipe_history(
      pipe_id=self.id,
      execution_start_ts=execution_start,
      strict=strict,
      succeeded=succeeded,
      in_memory=in_memory,
      passed_count=passed.shape[0],
      failed_count=failed.shape[0],
      quarantined_count=quarantined.shape[0]
    )

    # If the pipe failed, raise the EmptyPipe exception
    if not succeeded:
      raise EmptyPipe()
    # Otherwise, return remaining passed records
    return passed, failed, quarantined
  
  def add_metadata_columns(self, df: DataFrame) -> DataFrame:
    """Adds relevant metadata columns to the DataFrame before loading

    This method presumes the DataFrame carries its original metadata
    
    Args:
      df (DataFrame): the DataFrame before metadata columns
    
    Returns:
      df (DataFrame): the resulting DataFrame
    """
    id: str = str(uuid4())
    now: datetime = datetime.now(UTC)
    
    if self.table.quality.value == TableQuality.RAW.value:
      df: DataFrame = df.with_columns([
        pl.lit(id).alias('_raw_id'),
        pl.lit(now).alias('_ingested_ts')
      ])
    elif self.table.quality.value == TableQuality.CONFORMED.value:
      df: DataFrame = df.with_columns([
        pl.lit(id).alias('_conformed_id'),
        pl.lit(now).alias('_conformed_ts')
      ])
    elif self.table.quality.value == TableQuality.CANONICAL.value:
      df: DataFrame = df.with_columns([
        pl.lit(id).alias('_canonicalized_id'),
        pl.lit(now).alias('_created_ts'),
        pl.lit(now).alias('_modified_ts')
      ])
    else:
      raise TableQualityNotRecognized(self.table.quality.value)

    return df
  
  def conform_schema(self, df: DataFrame) -> DataFrame:
    """Conforms the DataFrame to the expected schema
    
    Args:
      df (DataFrame): the transformed, pre-conformed DataFrame
    
    Returns:
      df (DataFrame): the conformed DataFrame
    """
    df: DataFrame = self.add_metadata_columns(df)
    return df.select(*self.table.schema.polars.keys())

  def save(self, df: DataFrame) -> None:
    """Saves a DataFrame into the target Delta Table
    
    Args:
      df (DataFrame): the DataFrame to load
    """
    self.table.create_if_not_exists(
      table_path=self.table.table_path,
      schema=self.table.schema.deltalake
    )
    print(f'Loading {df.shape[0]} record(s) into {self.table.table_path}')

    if self.write_logic.value == WriteLogic.APPEND.value:
      self.table.append(df)
    elif self.write_logic.value == WriteLogic.OVERWRITE.value:
      self.table.overwrite(df)
    elif self.write_logic.value == WriteLogic.UPSERT.value:
      self.table.upsert(df)
    else:
      raise WriteLogicNotRecognized(self.write_logic)

  def quarantine(self, df: DataFrame) -> None:
    """Handles quarantined records from a save attempt

    The records get upserted into the corresponding quarantine table
    
    Args:
      df (DataFrame): the DataFrame of quarantined records
    """
    print(f'  - {df.shape[0]} record(s) got quarantined: {self.table.quarantine_path}')
    # Merge if the quarantine table exists
    # Otherwise, just append this time
    if DeltaTable.is_deltatable(self.table.quarantine_path):
      (df
        .write_delta(
          target=self.table.quarantine_path,
          mode='merge',
          delta_merge_options={
            'predicate': f's.{self.table.schema.failure_column} = t.{self.table.schema.failure_column}',
            'source_alias': 's',
            'target_alias': 't'
          }
        )
        .when_matched_update_all()
        .when_not_matched_insert_all()
        .execute()
      )
    else:
      df.write_delta(self.table.quarantine_path, mode='append')



from datetime import datetime
from polars import DataFrame
from typing import Literal, TypeAlias, TypedDict, Union


ExcelSpreadsheetEngine: TypeAlias = Literal['calamine', 'openpyxl', 'xlsx2csv']
RawPoltaData: TypeAlias = Union[DataFrame, dict, list[dict]]

class RawMetadata(TypedDict):
  """TypedDict to contain raw metadata for an ingest pipe"""
  _raw_id: str
  _ingested_ts: datetime
  _file_path: str
  _file_name: str
  _file_mod_ts: datetime



from dataclasses import dataclass, field
from datetime import datetime
from json import dumps
from os import makedirs, path
from polars import DataFrame
from typing import Optional

from polta.enums import ExportFormat, PipeType, WriteLogic
from polta.serializers import json
from polta.table import Table


@dataclass
class Exporter:
  """Executes an export of a Table
  
  Positional Args:
    table (Table): the main table for the export
    export_format (ExportFormat): how to save the data
  
  Optional Args:
    export_directory (str): where to save the data (default export volume)

  Initialized Fields:
    pipe_type (PipeType): what kind of pipe this is (i.e., EXPORTER)
    write_logic (Optional[WriteLogic]): how to save the data (i.e., None)
    exported_files (list[str]): the path of the exported files
  """
  table: Table
  export_format: ExportFormat
  export_directory: str = field(default_factory=lambda: '')

  pipe_type: PipeType = field(init=False)
  write_logic: Optional[WriteLogic] = field(init=False)
  exported_files: list[str] = field(init=False)

  def __post_init__(self) -> None:
    self.export_directory: str = self.export_directory or path.join(
      self.table.metastore.volumes_directory,
      'exports',
      self.table.domain,
      self.table.quality.value,
      self.table.name
    )
    makedirs(self.export_directory, exist_ok=True)
    self.pipe_type: PipeType = PipeType.EXPORTER
    self.write_logic = None
    self.exported_files: list[str] = []

  def get_dfs(self) -> dict[str, DataFrame]:
    """Retrieves the base DataFrame if possible, or it returns nothing

    Returns:
      dfs (dict[str, DataFrame]): the base DataFrame in a dict
    """
    df: DataFrame = self.table.get()
    return {} if df.is_empty() else {self.table.id: df}

  def transform(self, dfs: dict[str, DataFrame]) -> DataFrame:
    """Returns the target table DataFrame from dfs
    
    Args:
      dfs (dict[str, DataFrame]): the DataFrames to transform
    
    Returns:
      df (DataFrame): the resulting DataFrame
    """
    return dfs[self.table.id]

  def export(self, df: DataFrame) -> Optional[str]:
    """Exports the DataFrame to file storage

    Args:
      df (DataFrame): the DataFrame to export
    
    Returns:
      file_path (Optional[str]): if applicable, the resulting file_path
    """
    ts: str = datetime.now().strftime('%Y%m%d%H%M%S')
    file_name: str = f'{self.table.name}.{ts}.{self.export_format.value}'
    file_path: str = path.join(self.export_directory, file_name)
    if self.export_format.value == ExportFormat.CSV.value:
      self._to_csv(df, file_path)
    elif self.export_format.value == ExportFormat.JSON.value:
      self._to_json(df, file_path)
    else:
      raise NotImplementedError(f'Error: export format not implemented: {self.export_format.value}')
    self.exported_files.append(file_path)
    return file_path

  def _to_csv(self, df: DataFrame, file_path: str) -> None:
    """Exports the DataFrame to a CSV format

    Args:
      df (DataFrame): the DataFrame to export
      file_path (str): the target path of the file
    """
    df.write_csv(file_path)
  
  def _to_json(self, df: DataFrame, file_path: str) -> None:
    """Exports the DataFrame to a JSON format

    Args:
      df (DataFrame): the DataFrame to export
      file_path (str): the target path of the file
    """
    content: str = dumps(df.to_dicts(), default=json)
    open(file_path, 'w').write(content)



from enum import Enum


class CheckAction(Enum):
  """What to do if a check fails"""
  FAIL = 'fail'
  QUARANTINE = 'quarantine'

class DirectoryType(Enum):
  """Ingestion type"""
  SHALLOW = 'shallow'
  DATED = 'dated'

class ExportFormat(Enum):
  """Format of export files"""
  CSV = 'csv'
  JSON = 'json'

class PipeType(Enum):
  """Type of Polta Pipe logic"""
  EXPORTER = 'exporter'
  INGESTER = 'ingester'
  TRANSFORMER = 'transformer'

class RawFileType(Enum):
  """Format of raw files"""
  CSV = 'csv'
  EXCEL = 'xlsx'
  JSON = 'json'

class TableQuality(Enum):
  """The quality of the Delta Table"""
  RAW = 'raw'
  CONFORMED = 'conformed'
  CANONICAL = 'canonical'

class WriteLogic(Enum):
  """The method of saving data to a Delta Table"""
  APPEND = 'append'
  OVERWRITE = 'overwrite'
  UPSERT = 'upsert'



from datetime import date, datetime
from typing import Any


def json(obj: Any) -> str:
  """Makes a dict serializable with the json module
  
  Args:
    obj (Any): the value in the dict to serialize
  
  Returns:
    obj (str): the safely-formatted value
  """
  if isinstance(obj, (date, datetime)):
    return obj.isoformat()
  raise TypeError(f'Error: type {type(obj)} is not serializable')



import polars as pl

from polta.check import Check


check_value_in: Check = Check(
  name='value_in',
  description='Checks whether a column value is in a list',
  function=lambda df, column, values: (df
    .with_columns([
      pl.when(pl.col(column).is_in(values))
        .then(pl.lit(0))
        .otherwise(pl.lit(1))
        .alias(f'__{column}__value_in__')
    ])
  )
)



from dataclasses import dataclass, field
from deltalake import Field, Schema
from polars.datatypes import DataType

from polta.enums import TableQuality
from polta.maps import Maps
from polta.schemas.table import failed_test


@dataclass
class TableSchema:
  """Contains a schema with parameters indicating when the schema is effective
  
  Positional Args:
    raw_deltalake (Schema): the schema of the table (without metadata fields)
    quality (TableQuality): the quality of the table
  
  Initialized Fields:
    raw_polars (dict[str, DataType]): the raw schema as a polars dict
    metadata_fields (list[Field]): the table metadata fields based on quality
    metadata_columns (list[str]): the names of the metadata fields
    deltalake (Schema): the full table schema as a deltalake Schema
    polars (dict[str, DataType]): the full table schema as a polars dict
    raw_columns (list[str]): the names of the raw fields
    columns (list[str]): the names of the full table fields
  """
  raw_deltalake: Schema
  quality: TableQuality

  raw_polars: dict[str, DataType] = field(init=False)
  metadata_fields: list[Field] = field(init=False)
  deltalake: Schema = field(init=False)
  polars: dict[str, DataType] = field(init=False)
  quarantine: Schema = field(init=False)
  raw_columns: list[str] = field(init=False)
  metadata_columns: list[str] = field(init=False)
  columns: list[str] = field(init=False)
  failure_column: str = field(init=False)

  def __post_init__(self) -> None:
    # Create the schemas
    self.raw_polars: dict[str, DataType] = Maps \
      .deltalake_schema_to_polars_schema(self.raw_deltalake)
    self.metadata_fields: list[Field] = Maps \
      .quality_to_metadata_columns(self.quality)
    self.deltalake: Schema = Schema(
      self.metadata_fields + self.raw_deltalake.fields
    )
    self.polars: dict[str, DataType] = Maps \
      .deltalake_schema_to_polars_schema(self.deltalake)
    self.quarantine: dict[str, DataType] = Maps \
      .deltalake_schema_to_polars_schema(
        Schema(self.deltalake.fields + failed_test.fields)
      )
    self.failure_column: str = Maps.quality_to_failure_column(self.quality)

    # Store the columns
    self.raw_columns: list[str] = list(self.raw_polars.keys())
    self.metadata_columns: list[str] = list([f.name for f in self.metadata_fields])
    self.columns: list[str] = list(self.polars.keys())



from . import checks
from . import enums
from . import exceptions
from . import types
from . import udfs

from .check import Check
from .exporter import Exporter
from .ingester import Ingester
from .maps import Maps
from .metastore import Metastore
from .pipe import Pipe
from .pipeline import Pipeline
from .table_schema import TableSchema
from .table import Table
from .test import Test
from .transformer import Transformer


__all__ = [
  'Check',
  'checks',
  'enums',
  'exceptions',
  'Exporter',
  'Ingester',
  'Maps',
  'Metastore',
  'Pipe',
  'Pipeline',
  'Table',
  'TableSchema',
  'Test',
  'Transformer',
  'types',
  'udfs'
]
__author__ = 'JoshTG'
__license__ = 'MIT'



import polars as pl

from dataclasses import dataclass, field
from datetime import datetime, UTC
from deltalake import Field, Schema
from os import listdir, path
from polars import DataFrame
from polars.datatypes import DataType, List, String, Struct
from typing import Optional
from uuid import uuid4

from polta.enums import DirectoryType, PipeType, RawFileType, WriteLogic
from polta.exceptions import DirectoryTypeNotRecognized
from polta.maps import Maps
from polta.schemas.ingester import payload
from polta.table import Table
from polta.types import ExcelSpreadsheetEngine, RawMetadata
from polta.udfs import file_path_to_json, file_path_to_payload


@dataclass
class Ingester:
  """Dataclass for ingesting files into the target table
  
  Positional Args:
    table (Table): the target table for ingestion
    directory_type (DirectoryType): the kind of source directory to ingest
    raw_file_type (RawFileType): the format of the source files
  
  Optional Args:
    write_logic (WriteLogic): how to save the data (default APPEND)
  
  Initialized Fields:
    pipe_type (PipeType): what kind of pipe this is (i.e., INGESTER)
    simple_payload (bool): indicates whether the load is simple
    payload_schema (dict[str, DataType]): the polars fields for a simple ingestion
  """
  table: Table
  directory_type: DirectoryType
  raw_file_type: RawFileType
  write_logic: WriteLogic = field(default_factory=lambda: WriteLogic.APPEND)

  pipe_type: PipeType = field(init=False)
  simple_payload: bool = field(init=False)
  payload_schema: dict[str, DataType] = field(init=False)
  excel_engine: ExcelSpreadsheetEngine = field(default_factory=lambda: 'openpyxl')

  def __post_init__(self) -> None:
    self.pipe_type: PipeType = PipeType.INGESTER
    self.simple_payload: bool = self.table.schema.raw_deltalake.fields == payload.fields
    self.payload_schema: dict[str, DataType] = Maps.deltalake_schema_to_polars_schema(
      schema=Schema(self.table.schema.metadata_fields + payload.fields)
    )

  def get_dfs(self) -> dict[str, DataFrame]:
    """Ingests new files as DataFrames in a dict object
    
    Returns:
      dfs (dict[str, DataFrame]): the new DataFrames
    """
    df: DataFrame = self._get_metadata()
    df = self._filter_by_history(df)

    if df.is_empty():
      return {self.table.id: DataFrame([], self.table.schema.polars)}
    else:
      return {self.table.id: self._ingest_files(df)}

  def transform(self, dfs: dict[str, DataFrame]) -> DataFrame:
    """Returns the target table DataFrame from dfs
    
    Args:
      dfs (dict[str, DataFrame]): the DataFrames to transform
    
    Returns:
      df (DataFrame): the resulting DataFrame
    """
    return dfs[self.table.id]

  def export(self, df: DataFrame) -> Optional[str]:
    """Exports the DataFrame in a desired format

    This method is unused for ingesters

    Args:
      df (DataFrame): the DataFrame to export
    """
    return None

  def _get_file_paths(self) -> list[str]:
    """Retrieves a list of file paths based on ingestion parameters
        
    Returns:
      file_paths (list[str]): the resulting applicable file paths
    """
    if self.directory_type.value == DirectoryType.SHALLOW.value:
      return [
        path.join(self.table.ingestion_zone_path, f)
        for f in listdir(self.table.ingestion_zone_path)
      ]
    elif self.directory_type.value == DirectoryType.DATED.value:
      file_paths: list[str] = []
      for date_str in listdir(self.table.ingestion_zone_path):
        file_paths.extend([
          path.join(self.table.ingestion_zone_path, date_str, f)
          for f in listdir(path.join(self.table.ingestion_zone_path, date_str))
        ])
      return file_paths
    else:
      raise DirectoryTypeNotRecognized(self.directory_type)

  def _get_metadata(self) -> DataFrame:
    """Retrieves source file metadata as a DataFrame
        
    Returns:
      df (DataFrame): the metadata DataFrame
    """
    file_paths: list[str] = self._get_file_paths()
    metadata: list[RawMetadata] = [self._get_file_metadata(p) for p in file_paths]
    return DataFrame(metadata, schema=self.payload_schema)

  def _filter_by_history(self, metadata: list[RawMetadata]) -> DataFrame:
    """Removes files from ingestion attempt that have already been ingested
    
    Args:
      metadata (list[RawMetadata]): the file metadata to ingest
    
    Returns:
      file_paths (DataFrame): the resulting DataFrame object with the filtered paths
    """
    # Convert the file_paths field into a DataFrame
    paths: DataFrame = DataFrame(metadata, schema=self.payload_schema)

    # Retrieve the history from the target table
    hx: DataFrame = self.table.metastore.get_file_history(self.table.id)

    # Filter the paths by the temporary history DataFrame
    return (paths
      .join(hx, '_file_path', 'left')
      .filter(
        (pl.col('_file_mod_ts') > pl.col('_file_mod_ts_right')) |
        (pl.col('_file_mod_ts_right').is_null())
      )
      .drop('_file_mod_ts_right')
    )

  def _ingest_files(self, df: DataFrame) -> DataFrame:
    """Ingests files in the DataFrame according to file type / desired output
    
    Args:
      df (DataFrame): the files to load
    
    Returns:
      df (DataFrame): the ingested files
    """
    if self.simple_payload:
      df: DataFrame = self._run_simple_load(df)
    elif self.raw_file_type.value == RawFileType.CSV.value:
      df: DataFrame = self._run_csv_load(df)
    elif self.raw_file_type.value == RawFileType.JSON.value:
      df: DataFrame = self._run_json_load(df)
    elif self.raw_file_type.value == RawFileType.EXCEL.value:
      df: DataFrame = self._run_excel_load(df)
    else:
      raise NotImplementedError(self.raw_file_type)

    # Save the ingestion history
    self.table.metastore.write_file_history(self.table.id, df)

    return df

  def _get_file_metadata(self, file_path: str) -> RawMetadata:
    """Retrieves file metadata from a file

    Args:
      file_path (str): the path to the file
    
    Returns:
      raw_metadata (RawMetadata): the resulting raw metadata of the file
    """
    if not path.exists(file_path):
      raise FileNotFoundError()

    return RawMetadata(
      _raw_id=str(uuid4()),
      _ingested_ts=datetime.now(),
      _file_path=file_path,
      _file_name=path.basename(file_path),
      _file_mod_ts=datetime.fromtimestamp(path.getmtime(file_path), tz=UTC)
    )

  def _run_simple_load(self, df: DataFrame) -> DataFrame:
    """Retrieves the payload from the file path for each row
    
    Args:
      df (DataFrame): the data with metadata to load
    
    Returns:
      df (DataFrame): the resulting DataFrame
    """
    return (df
      .with_columns([
        pl.col('_file_path')
          .map_elements(file_path_to_payload, return_dtype=String)
          .alias('payload')
      ])
    )

  def _run_json_load(self, df: DataFrame) -> DataFrame:
    """Retrieves the payload values from the file path for each row
    
    Args:
      df (DataFrame): the data with metadata to load
    
    Returns:
      df (DataFrame): the resulting DataFrame
    """
    df: DataFrame = (df
      .with_columns([
        pl.col('_file_path')
          .map_elements(
            function=file_path_to_json,
            return_dtype=List(Struct(self.table.schema.raw_polars))
          )
          .alias('payload')
      ])
      .explode('payload')
      .with_columns([
        pl.col('payload').struct.field(f).alias(f)
        for f in self.table.schema.raw_polars.keys()
      ])
      .drop('payload')
    )
    return df
  
  def _run_csv_load(self, df: DataFrame) -> DataFrame:
    """Executes a CSV load against applicable files

    Args:
      df (DataFrame): the data with metadata to load

    Returns:
      df (DataFrame): the resulting DataFrame
    """
    out_df: Optional[DataFrame] = None
    for file_path in [r['_file_path'] for r in df.select('_file_path').to_dicts()]:
      csv_df: DataFrame = (pl
        .read_csv(
          source=file_path,
          schema=self.table.schema.raw_polars,
          try_parse_dates=True
        )
        .with_columns([pl.lit(file_path).alias('_file_path')])
      )
      if out_df is None:
        out_df: DataFrame = csv_df
      else:
        out_df: DataFrame = pl.concat([out_df, csv_df])
    out_df: DataFrame = out_df.join(df, '_file_path', 'inner')
    return out_df

  def _run_excel_load(self, df: DataFrame) -> DataFrame:
    """Executes an Excel load against applicable files

    Args:
      df (DataFrame): the data with metadata to load
    
    Returns:
      df (DataFrame): the resulting DataFrame
    """
    return (pl
      .read_excel(
        source=[r['_file_path'] for r in df.select('_file_path').to_dicts()],
        sheet_id=1,
        has_header=True,
        engine=self.excel_engine,
        include_file_paths='_file_path',
        drop_empty_rows=True,
        schema_overrides=self.table.schema.raw_polars
      )
      .join(df, '_file_path', 'inner')
    )



from json import load as json_load, loads
from typing import Any


def file_path_to_payload(file_path: str) -> str:
  """UDF to load a file as a string using open().read()
  
  Args:
    file_path (str): the path to the file
  
  Returns:
    payload (str): the resulting payload
  """
  return open(file_path, 'r').read()

def file_path_to_json(file_path: str) -> list[dict[str, Any]]:
  """UDF to load a file as a struct using json.load
  
  Args:
    file_path (str): the path to the file
  
  Returns:
    json (list[dict[str, Any]]): the file as a list of dict objects
  """
  return json_load(open(file_path, 'r'))

def string_to_struct(value: str) -> dict[str, Any]:
  """UDF to convert a string value to a struct value using json.loads
  
  Args:
    value (str): the value to convert
  
  Returns:
    struct (dict[str, Any]): the resulting struct"""
  return loads(value)



from .not_null_or_empty import check_not_null_or_empty
from .positive_int import check_positive_int
from .value_in import check_value_in


__all__ = [
  'check_not_null_or_empty',
  'check_positive_int',
  'check_value_in'
]



from deltalake import Field, Schema


file_history: Schema = Schema([
  Field('table_id', 'string'),
  Field('_file_path', 'string'),
  Field('_file_mod_ts', 'timestamp'),
  Field('_ingested_ts', 'timestamp')
])

pipe_history: Schema = Schema([
  Field('pipe_id', 'string'),
  Field('execution_start_ts', 'timestamp'),
  Field('execution_end_ts', 'timestamp'),
  Field('execution_duration', 'float'),
  Field('strict', 'boolean'),
  Field('succeeded', 'boolean'),
  Field('in_memory', 'boolean'),
  Field('total_count', 'long'),
  Field('passed_count', 'long'),
  Field('failed_count', 'long'),
  Field('quarantined_count', 'long')
])



from dataclasses import dataclass, field
from polars import DataFrame

from polta.pipe import Pipe


@dataclass
class Pipeline:
  """Simple dataclass for executing chains of pipes for an end product

  The pipe execution order:
    1. Raw
    2. Conformed
    3. Canonical
    4. Export
  
  Optional Args:
    raw_pipes (list[Pipe]): the raw pipes in the pipeline
    conformed_pipes (list[Pipe]): the conformed pipes in the pipeline
    canonical_pipes (list[Pipe]): the canonical pipes in the pipeline
    export_pipes (list[Pipe]): the export pipes in the pipeline
  """
  raw_pipes: list[Pipe] = field(default_factory=lambda: [])
  conformed_pipes: list[Pipe] = field(default_factory=lambda: [])
  canonical_pipes: list[Pipe] = field(default_factory=lambda: [])
  export_pipes: list[Pipe] = field(default_factory=lambda: [])

  def execute(self, in_memory: bool = False, skip_exports: bool = False) -> None:
    """Executes all available pipes in order of layer
    
    Args:
      in_memory (bool): indicates whether to run without saving (default False)
      skip_exports (bool): indicates whether to skip the export layer (default False)
    """
    if in_memory:
      self._in_memory_execute(skip_exports)
    else:
      self._standard_execute(skip_exports)

  def _standard_execute(self, skip_exports: bool = False) -> None:
    """Executes all available pipes in order of layer and saves results
    
    Args:
      skip_exports (bool): indicates whether to skip the export layer (default False)
    """
    for pipe in self.raw_pipes:
      pipe.execute()
    for pipe in self.conformed_pipes:
      pipe.execute()
    for pipe in self.canonical_pipes:
      pipe.execute()
    if not skip_exports:
      for pipe in self.export_pipes:
        pipe.execute()

  def _in_memory_execute(self, skip_exports: bool = False) -> None:
    """Executes all available pipes in order of layer without saving to the metastore

    Args:
      skip_exports (bool): indicates whether to skip the export layer (default False)    
    """
    dfs: dict[str, DataFrame] = {}

    for pipe in self.raw_pipes:
      passed, _, _ = pipe.execute(dfs, in_memory=True)
      dfs[pipe.table.id] = passed
    for pipe in self.conformed_pipes:
      passed, _, _ = pipe.execute(dfs, in_memory=True)
      dfs[pipe.table.id] = passed
    for pipe in self.canonical_pipes:
      passed, _, _ = pipe.execute(dfs, in_memory=True)
      dfs[pipe.table.id] = passed
    if not skip_exports:
      for pipe in self.export_pipes:
        passed, _, _ = pipe.execute(dfs, in_memory=True)
        dfs[pipe.table.id] = passed



from dataclasses import dataclass, field
from polars import DataFrame

from polta.check import Check
from polta.enums import CheckAction


@dataclass
class Test:
  check: Check
  column: str
  check_action: CheckAction
  kwargs: dict = field(default_factory=lambda: {})

  result_column: str = field(init=False)

  def __post_init__(self) -> None:
    self.result_column: str = self.check.build_result_column(self.column)

  def run(self, df: DataFrame) -> DataFrame:
    if self.check.simple_function:
      return self.check.function(df, self.column)
    else:
      return self.check.function(df, self.column, **self.kwargs)



from dataclasses import dataclass, field
from polars import DataFrame, SQLContext
from types import FunctionType
from typing import Optional, Union

from polta.enums import PipeType, WriteLogic
from polta.exceptions import IncompatibleTransformLogic
from polta.table import Table


@dataclass
class Transformer:
  """Contains transformation logic to be used in a Pipe
  
  Positional Args:
    table (Table): the target Table
    load_logic (FunctionType): a method to load the source DataFrames
    transform_logic (Union[FunctionType, str]): a method/query to transform the DataFrames
  
  Optional Args:
    write_logic (WriteLogic): how to write to a Table (default APPEND)
  
  Initialized fields:
    pipe_type (PipeType): the type of pipe this is (i.e., TRANSFORMER)
  """
  table: Table
  load_logic: FunctionType
  transform_logic: Union[FunctionType, str]
  write_logic: WriteLogic = field(default_factory=lambda: WriteLogic.APPEND)

  pipe_type: PipeType = field(init=False)

  def __post_init__(self) -> None:
    self.pipe_type: PipeType = PipeType.TRANSFORMER

  def get_dfs(self) -> dict[str, DataFrame]:
    """Executes the load_logic callable to return source DataFrames

    Returns:
      dfs (dict[str, DataFrame]): the source DataFrames
    """
    return self.load_logic()

  def transform(self, dfs: dict[str, DataFrame]) -> DataFrame:
    """Applies the transform_logic callable to the DataFrames

    Args:
      dfs (dict[str, DataFrame]): the DataFrames to transform

    Returns:
      df (DataFrame): the transformed DataFrame
    """
    if isinstance(self.transform_logic, FunctionType):
      return self.transform_logic(dfs)
    elif isinstance(self.transform_logic, str):
      return SQLContext(frames=dfs).execute(
        query=self.transform_logic,
        eager=True
      )
    else:
      raise IncompatibleTransformLogic(self.transform_logic)
  
  def export(self, df: DataFrame) -> Optional[str]:
    """Exports the DataFrame in a desired format

    This method is unused for transformers

    Args:
      df (DataFrame): the DataFrame to export
    """
    return None



import polars as pl

from polta.check import Check


check_positive_int: Check = Check(
  name='positive_int',
  description='Checks whether column values are positive integers',
  function=lambda df, column: (df
    .with_columns([
      pl.when(pl.col(column) >= 0)
        .then(pl.lit(1))
        .otherwise(pl.lit(0))
        .alias(f'__{column}__positive_int__')
    ])
  )
)



from deltalake import Field, Schema


raw_metadata: Schema = Schema([
  Field('_raw_id', 'string'),
  Field('_ingested_ts', 'timestamp'),
  Field('_file_path', 'string'),
  Field('_file_name', 'string'),
  Field('_file_mod_ts', 'timestamp')
])

conformed_metadata: Schema = Schema([
  Field('_raw_id', 'string'),
  Field('_conformed_id', 'string'),
  Field('_conformed_ts', 'timestamp'),
  Field('_ingested_ts', 'timestamp'),
  Field('_file_path', 'string'),
  Field('_file_name', 'string'),
  Field('_file_mod_ts', 'timestamp')
])

canonical_metadata: Schema = Schema([
  Field('_raw_id', 'string'),
  Field('_conformed_id', 'string'),
  Field('_canonicalized_id', 'string'),
  Field('_created_ts', 'timestamp'),
  Field('_modified_ts', 'timestamp')
])

failed_test: Schema = Schema([
  Field('failed_test', 'string')
])



import polars as pl

from polta.check import Check


check_not_null_or_empty: Check = Check(
  name='not_null_or_empty',
  description='Checks whether column values are not null or empty',
  function=lambda df, column: (df
    .with_columns([
      pl.when(pl.col(column).is_not_null(), pl.col(column) != '')
        .then(pl.lit(1))
        .otherwise(pl.lit(0))
        .alias(f'__{column}__not_null_or_empty__')
    ])
  )
)



from typing import Any


class DataTypeNotRecognized(Exception):
  """Raise when the data type is not recognized by system"""
  def __init__(self, data_type: str) -> None:
    self.message: str = data_type
    super().__init__(self.message)

class DirectoryTypeNotRecognized(Exception):
  """Raise when the directory type is not recognized by system"""
  def __init__(self, directory_type: Any) -> None:
    self.message: str = directory_type
    super().__init__(self.message)

class DomainDoesNotExist(Exception):
  """Raise when a domain retrieval is attempted but it does not exist"""
  def __init__(self, domain: str) -> None:
    self.message: str = f'Unrecognized domain {domain}'
    super().__init__(self.message)

class EmptyPipe(Exception):
  """Raise when the pipe should load data but did not"""
  def __init__(self) -> None:
    self.message: str = f'Pipe executed in strict mode but did not load data'
    super().__init__(self.message)

class IncompatibleTransformLogic(Exception):
  """Raise when the transform logic is not compatible with system"""
  def __init__(self, transform_logic: Any) -> None:
    self.message: str = f'Incompatible transform logic of type {type(transform_logic)}'
    super().__init__(self.message)

class PoltaDataFormatNotRecognized(Exception):
  """Raise when the Polta data format is not recognized by system"""
  def __init__(self, format: type) -> None:
    self.message: str = f'Unrecognized instance type {format}'
    super().__init__(self.message)

class TableQualityNotRecognized(Exception):
  """Raise when the table quality is not recognized by system"""
  def __init__(self, quality: Any) -> None:
    self.message: str = f'Unrecognized table quality {quality}'
    super().__init__(self.message)

class WriteLogicNotRecognized(Exception):
  """Raise when the write logic is not recognized by system"""
  def __init__(self, write_logic: Any) -> None:
    self.message: str = f'Unrecognized write logic {str(write_logic)}'
    super().__init__(self.message)
